import Foundation
import SQLite

extension Metadata.Value {

    private static let table = Table("metadata_values")
    
    private static let columnRowId = Expression<Int>("ROW_ID")

    private static let columnKeyId = Expression<Int?>("key_id")

    private static let columnObjectId = Expression<Int?>("object_id")
    
    private static let columnValueType = Expression<Int>("value_type")
    
    private static let columnStringValue = Expression<String?>("string_value")
    
    private static let columnNumericalValue = Expression<Double?>("numerical_value")
    
    private static let columnDateValue = Expression<Double?>("date_value")
    
    private static let columnDataValue = Expression<Data?>("data_value")
    
    private static func readAll(in database: Connection) throws -> [Self] {
        try database.prepare(table).map(from)
    }

    static func metadata(for workoutId: Int, in database: Connection) throws -> [Self] {
        try database.prepare(table.filter(columnObjectId == workoutId)).map(from)
    }

    static func metadata(for workoutId: Int, in database: Connection) throws -> [(keyId: Int, value: Self)] {
        try database.prepare(table.filter(columnObjectId == workoutId)).compactMap { row in
            guard let keyId = row[columnKeyId] else {
                print("Found 'key_id == NULL' for metadata value of workout \(workoutId)")
                return nil
            }
            return (keyId, from(row: row))
        }
    }

    static func createTable(in database: Connection) throws {
        //try database.execute("CREATE TABLE metadata_values (ROWID INTEGER PRIMARY KEY AUTOINCREMENT, key_id INTEGER, object_id INTEGER, value_type INTEGER NOT NULL DEFAULT 0, string_value TEXT, numerical_value REAL, date_value REAL, data_value BLOB)")
        try database.run(table.create { table in
            table.column(columnRowId, primaryKey: .autoincrement)
            table.column(columnKeyId)
            table.column(columnObjectId)
            table.column(columnValueType, defaultValue: 0)
            table.column(columnStringValue)
            table.column(columnNumericalValue)
            table.column(columnDateValue)
            table.column(columnDataValue)
        })
    }

    static func insert(_ element: Self, of workoutId: Int, for keyId: Int, in database: Connection) throws {
        try database.run(table.insert(
            columnKeyId <- keyId,
            columnObjectId <- workoutId,
            columnValueType <- element.valueType.rawValue,
            columnStringValue <- element.stringValue,
            columnNumericalValue <- element.numericalValue,
            columnDateValue <- element.dateValue,
            columnDataValue <- element.dataValue
        ))
    }
}

private extension Metadata.Value {

    var stringValue: String? {
        if case let .string(value) = self {
            return value
        }
        if case let .numerical(_, unit) = self {
            return unit
        }
        return nil
    }

    var numericalValue: Double? {
        if case let .number(value) = self {
            return value
        }
        if case let .numerical(value: value, unit: _) = self {
            return value
        }
        return nil
    }

    var dateValue: Double? {
        if case let .date(value: date) = self {
            return date.timeIntervalSinceReferenceDate
        }
        return nil
    }

    var dataValue: Data? {
        if case let .data(data) = self {
            return data
        }
        return nil
    }
}


extension Metadata.Value {

    static func from(row: Row) -> Self {
        let valueType = ValueType(rawValue: row[columnValueType])!
        switch valueType {
        case .string:
            return .string(value: row[columnStringValue]!)
        case .number:
            return .number(value: row[columnNumericalValue]!)
        case .date:
            return .date(value: .init(timeIntervalSinceReferenceDate: row[columnDateValue]!))
        case .numerical:
            return .numerical(value: row[columnNumericalValue]!, unit: row[columnStringValue]!)
        case .data:
            return .data(value: row[columnDataValue]!)
        }
    }
}
