import Foundation
import SQLite

extension Workout {

    private static let table = Table("workouts")
    
    // INTEGER PRIMARY KEY AUTOINCREMENT
    private static let columnDataId = Expression<Int>("data_id")
    
    // REAL
    private static let columnTotalDistance = Expression<Double?>("total_distance")
    
    // INTEGER
    private static let columnGoalType = Expression<Int?>("goal_type")
    
    // REAL
    private static let columnGoal = Expression<Double?>("goal")
    
    // INTEGER
    private static let columnCondenserVersion = Expression<Int?>("condenser_version")
    
    // REAL
    private static let columnCondenserDate = Expression<Double?>("condenser_date")

    static func readAll(in database: Connection) throws -> [Workout] {
        let metadataKeys = try Metadata.allKeys(in: database)

        return try database.prepare(table).map { row in
            let id = row[columnDataId]

            let events = try WorkoutEvent.events(for: id, in: database)
            let activities = try WorkoutActivity.activities(for: id, in: database)
            let metadata = try Metadata.metadata(for: id, in: database, keyMap: metadataKeys)
            return .init(
                id: id,
                totalDistance: row[columnTotalDistance],
                goalType: row[columnGoalType],
                goal:  row[columnGoal],
                condenserVersion: row[columnCondenserVersion],
                condenserDate: row[columnCondenserDate].map { Date.init(timeIntervalSinceReferenceDate: $0) },
                events: events,
                activities: activities,
                metadata: metadata)
        }
    }

    static func createTable(in database: Database) throws {
        try database.run(table.create { t in
            t.column(columnDataId, primaryKey: .autoincrement)
            t.column(columnTotalDistance)
            t.column(columnGoalType)
            t.column(columnGoal)
            t.column(columnCondenserVersion)
            t.column(columnCondenserDate)
        })
        // try database.execute("CREATE TABLE workouts (data_id INTEGER PRIMARY KEY AUTOINCREMENT, total_distance REAL, goal_type INTEGER, goal REAL, condenser_version INTEGER, condenser_date REAL)")
    }

    func insert(in database: Database) throws {
        try Workout.insert(self, in: database)
    }

    private static func insert(_ element: Workout, in database: Database) throws {
        let rowid = try database.run(table.insert(
            columnTotalDistance <- element.totalDistance,
            columnGoalType <- element.goalType,
            columnGoalType <- element.goalType,
            columnGoal <- element.goal,
            columnCondenserVersion <- element.condenserVersion,
            columnCondenserDate <- element.condenserDate?.timeIntervalSinceReferenceDate)
        )
        let dataId = Int(rowid)
        for event in element.events {
            try event.insert(in: database, dataId: dataId)
        }
        for activity in element.activities {
            try activity.insert(in: database, dataId: dataId)
        }
        for (key, value) in element.metadata {
            try Metadata.insert(value, for: key, of: dataId, in: database)
        }
    }
}

